package net.es.oscars.policy.axis2.service;

import java.util.Arrays;

import org.apache.axiom.om.OMElement;
import org.apache.axis2.util.XMLUtils;
import org.apache.log4j.Logger;
import org.w3c.dom.Element;

import net.es.oscars.policy.axis2.types.CheckPolicyRequest;
import net.es.oscars.policy.axis2.types.CheckPolicyRequestType;
import net.es.oscars.policy.axis2.types.CheckPolicyResponse;
import net.es.oscars.policy.axis2.types.CheckPolicyResponseType;
import net.es.oscars.policy.axis2.types.RequesterType;
import net.es.oscars.policy.axis2.types.ResourceType;
import net.es.oscars.policy.axis2.util.RequesterComparator;
import net.es.oscars.policy.client.DOMRequester;
import net.es.oscars.policy.rmi.PolicyRmiClient;

/**
 * PolicyAdapter class converts Axis2 web service requests into RMI requests 
 * and passes them to the core. It also processes the responses and converts 
 * them from the format returned by RMI to an Axis2 type to be returned as a 
 * web service message. The RMI interface uses W3C DOM elements so many of the 
 * conversions involve changing Axis2 OMElements to DOM elements. 
 * 
 * @author Andrew Lake (alake@internet2.edu)
 */
public class PolicyAdapter {
    private Logger log;
    
    public PolicyAdapter(){
        this.log = Logger.getLogger(this.getClass());
    }
    
    /**
     * Converts an Axis2 checkPolicy call to an RMI checkPolicy call. 
     * 
     * @param request the Axis2 request 
     * @return an Axis2 CheckPolicyResponse that can be returned by the skeleton
     * @throws Exception
     */
    public CheckPolicyResponse checkPolicy(CheckPolicyRequest request) throws Exception{
        PolicyRmiClient rmiClient = new PolicyRmiClient();
        CheckPolicyRequestType axis2Params = request.getCheckPolicyRequest();
        
        net.es.oscars.policy.client.CheckPolicyRequest<DOMRequester,Element[]> rmiRequest = 
            new net.es.oscars.policy.client.CheckPolicyRequest<DOMRequester,Element[]>();
        
        //Convert requesters
        RequesterType[] axis2Requesters = axis2Params.getRequesters().getRequester();
        //sort by sequence number
        Arrays.sort(axis2Requesters, new RequesterComparator());
        for(RequesterType axis2Requester : axis2Requesters){
            DOMRequester requester = new DOMRequester();
            //Set Subject
            requester.setSubject(XMLUtils.toDOM(axis2Requester.getExtraElement()));
            //Set Subject Authentication
            OMElement axis2SubjAuth = null;
            if(axis2Requester.getSubjectAuthentication() != null){
                axis2SubjAuth = axis2Requester.getSubjectAuthentication().getExtraElement();
            }
            if(axis2SubjAuth != null){
                requester.setSubjectAuthentication(XMLUtils.toDOM(axis2SubjAuth));
            }
            //Set Assertions
            OMElement[] axis2Asserts = null;
            if(axis2Requester.getRequesterTypeSequence_type0() != null){
                axis2Asserts = axis2Requester.getRequesterTypeSequence_type0().getExtraElement();
            }
            if(axis2Asserts != null){
                for(OMElement axis2Assert : axis2Requester.getRequesterTypeSequence_type0().getExtraElement()){
                       requester.getAssertions().add(XMLUtils.toDOM(axis2Assert));
                }
            }
            rmiRequest.getRequesters().add(requester);
        }
        
        //Convert action
        rmiRequest.setAction(axis2Params.getAction().toString());
        
        //Convert resources
        if(axis2Params.getResource() != null){
            for(ResourceType resource : axis2Params.getResource()){
                if(resource.getExtraElement() == null){ 
                    continue;
                }
                
                Element[] elements = new Element[resource.getExtraElement().length];
                for(int i = 0; i < resource.getExtraElement().length; i++){
                    elements[i] = XMLUtils.toDOM(resource.getExtraElement()[i]);
                }
                rmiRequest.getResources().add(elements);
            }
        }
        
        //send request
        rmiClient.init();
        net.es.oscars.policy.client.CheckPolicyResponse rmiResponse = 
            rmiClient.checkPolicy(rmiRequest);
        this.log.info(rmiResponse.getReason());
        
        //convert response to Axis2
        CheckPolicyResponse responseWrapper = new CheckPolicyResponse();
        CheckPolicyResponseType response = new CheckPolicyResponseType();
        response.setAllow(rmiResponse.getAllow());
        response.setReason(rmiResponse.getReason());
        responseWrapper.setCheckPolicyResponse(response);
        
        return responseWrapper;
    }
}
